'''
Version and license information.
'''

__all__ = ['__version__', '__versiondate__', '__license__']

__version__ = '1.0.0'
__versiondate__ = '2023-01-31'
__license__ = f'HPVsim {__version__} ({__versiondate__}) — © 2023 by IDM'
